"""
This module rasters/accumulates the spikes generated by spike generation module into a lower clock-rate signal.
Please note that this module is NOT implemented in hardware as it is.
Instead, the input spike registers of the SNN core `ISPKREG` accumulates spikes in the same way as this module does.
It's important to use the same dt that this module generates to make sure to have a valid simulation!

``dt = rate_downsample_factor / AUDIO_SAMPLING_RATE_PDM``
"""

# - Rockpool imports
from rockpool.nn.modules.module import Module
from rockpool.parameters import SimulationParameter

import numpy as np


from typing import Union, Tuple
from rockpool.typehints import P_int, P_float, P_ndarray


# list of modules exported
__all__ = ["Raster"]


class Raster(Module):
    def __init__(
        self,
        shape: Union[int, Tuple[int]],
        fs: float,
        rate_downsample_factor: int = 2**6,
        max_num_spikes: int = 15,
    ):
        """this class rasters the high clock-rate spikes produced by the spike generation module into a signal
        with a lower clock rate.

        Args:
            shape (Union[int, Tuple[int]]): shape of the input spikes. If the input is a single channel, then the shape is an integer. Otherwise, it is a tuple of integers.
            fs (float, optional): clock rate of the input spikes (around 48K ~ 50K).
            rate_downsample (int, optional): How much the spike rate is going to be reduced. Defaults to 2**6 (i.e. moving from around 50K down to around 800)
            max_num_spikes (int, optional): Maximum number of spikes transferred to the core SNN. Defaults to 15.

        """
        super().__init__(shape=shape)

        self.rate_downsample_factor: P_int = SimulationParameter(rate_downsample_factor)
        self.max_num_spikes: P_int = SimulationParameter(max_num_spikes)
        self.fs: P_float = SimulationParameter(fs)

    def evolve(
        self, spikes_in: np.ndarray, record: bool = False, *args, **kwargs
    ) -> np.ndarray:
        """this modules rasters the input spikes into a signal with lower clock rate.

        Args:
            spikes_in (np.ndarray): `T x num_channel` input spikes received from spike generation module.

        Returns:
            np.ndarray: rastered spikes.
        """
        # check the dimension
        if spikes_in.ndim == 1:
            spikes_in = spikes_in.reshape(-1, 1)

        # accumulate the spikes
        acc_spikes_in = np.cumsum(spikes_in, axis=0)

        # index set to sample the accumulated spikes
        if spikes_in.shape[0] % self.rate_downsample_factor == 0:
            index_set = np.arange(
                self.rate_downsample_factor - 1,
                spikes_in.shape[0],
                step=self.rate_downsample_factor,
            )
        else:
            index_set = np.concatenate(
                [
                    np.arange(
                        self.rate_downsample_factor - 1,
                        spikes_in.shape[0],
                        step=self.rate_downsample_factor,
                    ),
                    [spikes_in.shape[0] - 1],
                ]
            )

        rastered_spikes = acc_spikes_in[index_set, :]

        # compute the accumulated spikes within the sampling period
        rastered_spikes[1:, :] -= rastered_spikes[:-1, :]

        # truncate the number of spikes
        rastered_spikes[rastered_spikes > self.max_num_spikes] = self.max_num_spikes

        # reshape the spikes in case the signal has single channel
        rastered_spikes = rastered_spikes.squeeze()

        return rastered_spikes, self.state(), {}

    def __call__(self, *args, **kwargs) -> np.ndarray:
        """This is the same as `evolve` function.

        Returns:
            np.ndarray: rastered spikes.
        """
        return self.evolve(*args, **kwargs)

    def _info(self) -> str:
        string = (
            "This module rasters/accumulates the spikes received from spike generation module into a lower clock-rate signal.\n"
            + "parameters:\n"
            + f"clock rate downsampling factor: {self.rate_downsample_factor}\n"
            + f"maximum number of spikes within a raster period/interval: {self.max_num_spikes}\n"
        )
        return string
